# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import aqt
import aqt.forms
import aqt.operations
from anki.collection import OpChanges
from anki.lang import without_unicode_isolation
from anki.models import NotetypeDict
from aqt import AnkiQt, gui_hooks
from aqt.operations.notetype import update_notetype_legacy
from aqt.qt import *
from aqt.schema_change_tracker import ChangeTracker
from aqt.utils import (
    HelpPage,
    add_close_shortcut,
    askUser,
    disable_help_button,
    getOnlyText,
    openHelp,
    show_warning,
    tooltip,
    tr,
)


class FieldDialog(QDialog):
    def __init__(
        self,
        mw: AnkiQt,
        nt: NotetypeDict,
        parent: QWidget | None = None,
        open_at: int = 0,
    ) -> None:
        QDialog.__init__(self, parent or mw)
        mw.garbage_collect_on_dialog_finish(self)
        self.mw = mw
        self.col = self.mw.col
        self.mm = self.mw.col.models
        self.model = nt
        self.mm._remove_from_cache(self.model["id"])
        self.change_tracker = ChangeTracker(self.mw)
        self.webview = None

        self.form = aqt.forms.fields.Ui_Dialog()
        self.form.setupUi(self)

        self.setWindowTitle(
            without_unicode_isolation(tr.fields_fields_for(val=self.model["name"]))
        )

        add_close_shortcut(self)
        disable_help_button(self)
        help_button = self.form.buttonBox.button(QDialogButtonBox.StandardButton.Help)
        assert help_button is not None
        help_button.setAutoDefault(False)

        cancel_button = self.form.buttonBox.button(
            QDialogButtonBox.StandardButton.Cancel
        )
        assert cancel_button is not None
        cancel_button.setAutoDefault(False)

        save_button = self.form.buttonBox.button(QDialogButtonBox.StandardButton.Save)
        assert save_button is not None
        save_button.setAutoDefault(False)

        self.currentIdx: int | None = None
        self.fillFields()
        self.setupSignals()
        self.form.fieldList.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self.form.fieldList.dropEvent = self.onDrop  # type: ignore[assignment]
        self.form.fieldList.setCurrentRow(open_at)
        self.exec()

    def _on_bridge_cmd(self, cmd: str) -> bool:
        return False

    ##########################################################################

    def fillFields(self) -> None:
        self.currentIdx = None
        self.form.fieldList.clear()
        for c, f in enumerate(self.model["flds"]):
            self.form.fieldList.addItem(f"{c + 1}: {f['name']}")

    def setupSignals(self) -> None:
        f = self.form
        qconnect(f.fieldList.currentRowChanged, self.onRowChange)
        qconnect(f.fieldAdd.clicked, self.onAdd)
        qconnect(f.fieldDelete.clicked, self.onDelete)
        qconnect(f.fieldRename.clicked, self.onRename)
        qconnect(f.fieldPosition.clicked, self.onPosition)
        qconnect(f.sortField.clicked, self.onSortField)
        qconnect(f.buttonBox.helpRequested, self.onHelp)

    def onDrop(self, ev: QDropEvent) -> None:
        fieldList = self.form.fieldList
        indicatorPos = fieldList.dropIndicatorPosition()
        if qtmajor == 5:
            pos = ev.pos()  # type: ignore
        else:
            pos = ev.position().toPoint()
        dropPos = fieldList.indexAt(pos).row()
        idx = self.currentIdx
        if dropPos == idx:
            return
        if (
            indicatorPos == QAbstractItemView.DropIndicatorPosition.OnViewport
        ):  # to bottom.
            movePos = fieldList.count() - 1
        elif indicatorPos == QAbstractItemView.DropIndicatorPosition.AboveItem:
            movePos = dropPos
        elif indicatorPos == QAbstractItemView.DropIndicatorPosition.BelowItem:
            movePos = dropPos + 1
        else:
            # for pylint
            return
        # the item in idx is removed thus subtract 1.
        assert idx is not None
        if idx < dropPos:
            movePos -= 1
        self.moveField(movePos + 1)  # convert to 1 based.

    def onRowChange(self, idx: int) -> None:
        if idx == -1:
            return
        self.saveField()
        self.loadField(idx)

    def _uniqueName(
        self, prompt: str, ignoreOrd: int | None = None, old: str = ""
    ) -> str | None:
        txt = getOnlyText(prompt, default=old).replace('"', "").strip()
        if not txt:
            return None
        if txt[0] in "#^/":
            show_warning(tr.fields_name_first_letter_not_valid())
            return None
        for letter in """:{"}""":
            if letter in txt:
                show_warning(tr.fields_name_invalid_letter())
                return None
        for f in self.model["flds"]:
            if ignoreOrd is not None and f["ord"] == ignoreOrd:
                continue
            if f["name"] == txt:
                show_warning(tr.fields_that_field_name_is_already_used())
                return None
        return txt

    def onRename(self) -> None:
        if self.currentIdx is None:
            return

        idx = self.currentIdx
        f = self.model["flds"][idx]
        name = self._uniqueName(tr.actions_new_name(), self.currentIdx, f["name"])
        if not name:
            return

        old_name = f["name"]
        self.change_tracker.mark_basic()
        self.mm.rename_field(self.model, f, name)
        gui_hooks.fields_did_rename_field(self, f, old_name)

        self.saveField()
        self.fillFields()
        self.form.fieldList.setCurrentRow(idx)

    def onAdd(self) -> None:
        name = self._uniqueName(tr.fields_field_name())
        if not name:
            return
        if not self.change_tracker.mark_schema():
            return
        self.saveField()
        f = self.mm.new_field(name)
        self.mm.add_field(self.model, f)
        gui_hooks.fields_did_add_field(self, f)

        self.fillFields()
        self.form.fieldList.setCurrentRow(len(self.model["flds"]) - 1)

    def onDelete(self) -> None:
        if len(self.model["flds"]) < 2:
            show_warning(tr.fields_notes_require_at_least_one_field())
            return
        field = self.model["flds"][self.form.fieldList.currentRow()]
        if field["preventDeletion"]:
            show_warning(tr.fields_field_is_required())
            return
        count = self.mm.use_count(self.model)
        c = tr.browsing_note_count(count=count)
        if not askUser(tr.fields_delete_field_from(val=c)):
            return
        if not self.change_tracker.mark_schema():
            return
        self.mm.remove_field(self.model, field)
        gui_hooks.fields_did_delete_field(self, field)

        self.fillFields()
        self.form.fieldList.setCurrentRow(0)

    def onPosition(self, delta: int = -1) -> None:
        idx = self.currentIdx
        assert idx is not None
        l = len(self.model["flds"])
        txt = getOnlyText(tr.fields_new_position_1(val=l), default=str(idx + 1))
        if not txt:
            return
        try:
            pos = int(txt)
        except ValueError:
            return
        if not 0 < pos <= l:
            return
        self.moveField(pos)

    def onSortField(self) -> None:
        if not self.change_tracker.mark_schema():
            return
        # don't allow user to disable; it makes no sense
        self.form.sortField.setChecked(True)
        self.mm.set_sort_index(self.model, self.form.fieldList.currentRow())

    def moveField(self, pos: int) -> None:
        if not self.change_tracker.mark_schema():
            return
        self.saveField()
        f = self.model["flds"][self.currentIdx]
        self.mm.reposition_field(self.model, f, pos - 1)
        self.fillFields()
        self.form.fieldList.setCurrentRow(pos - 1)

    def loadField(self, idx: int) -> None:
        self.currentIdx = idx
        fld = self.model["flds"][idx]
        f = self.form
        f.fontFamily.setCurrentFont(QFont(fld["font"]))
        f.fontSize.setValue(fld["size"])
        f.sortField.setChecked(self.model["sortf"] == fld["ord"])
        f.rtl.setChecked(fld["rtl"])
        f.plainTextByDefault.setChecked(fld["plainText"])
        f.collapseByDefault.setChecked(fld["collapsed"])
        f.excludeFromSearch.setChecked(fld["excludeFromSearch"])
        f.fieldDescription.setText(fld.get("description", ""))

    def saveField(self) -> None:
        # not initialized yet?
        if self.currentIdx is None:
            return
        idx = self.currentIdx
        fld = self.model["flds"][idx]
        f = self.form
        font = f.fontFamily.currentFont().family()
        if fld["font"] != font:
            fld["font"] = font
            self.change_tracker.mark_basic()
        size = f.fontSize.value()
        if fld["size"] != size:
            fld["size"] = size
            self.change_tracker.mark_basic()
        rtl = f.rtl.isChecked()
        if fld["rtl"] != rtl:
            fld["rtl"] = rtl
            self.change_tracker.mark_basic()
        plain_text = f.plainTextByDefault.isChecked()
        if fld["plainText"] != plain_text:
            fld["plainText"] = plain_text
            self.change_tracker.mark_basic()
        collapsed = f.collapseByDefault.isChecked()
        if fld["collapsed"] != collapsed:
            fld["collapsed"] = collapsed
            self.change_tracker.mark_basic()
        exclude_from_search = f.excludeFromSearch.isChecked()
        if fld["excludeFromSearch"] != exclude_from_search:
            fld["excludeFromSearch"] = exclude_from_search
            self.change_tracker.mark_basic()
        desc = f.fieldDescription.text()
        if fld.get("description", "") != desc:
            fld["description"] = desc
            self.change_tracker.mark_basic()

    def reject(self) -> None:
        if self.webview:
            self.webview.cleanup()
            self.webview = None

        if self.change_tracker.changed():
            if not askUser(tr.card_templates_discard_changes()):
                return

        QDialog.reject(self)

    def accept(self) -> None:
        self.saveField()

        def on_done(changes: OpChanges) -> None:
            tooltip(tr.card_templates_changes_saved(), parent=self.parentWidget())
            QDialog.accept(self)

        update_notetype_legacy(
            parent=self.mw, notetype=self.model, skip_checks=True
        ).success(on_done).run_in_background()

    def onHelp(self) -> None:
        openHelp(HelpPage.CUSTOMIZING_FIELDS)



# Form implementation generated from reading ui file 'qt/aqt/forms/customstudy.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(332, 380)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.radioAhead = QtWidgets.QRadioButton(parent=Dialog)
        self.radioAhead.setObjectName("radioAhead")
        self.gridLayout.addWidget(self.radioAhead, 3, 0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.radioForgot = QtWidgets.QRadioButton(parent=Dialog)
        self.radioForgot.setObjectName("radioForgot")
        self.gridLayout.addWidget(self.radioForgot, 2, 0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.radioNew = QtWidgets.QRadioButton(parent=Dialog)
        self.radioNew.setObjectName("radioNew")
        self.gridLayout.addWidget(self.radioNew, 0, 0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.radioRev = QtWidgets.QRadioButton(parent=Dialog)
        self.radioRev.setObjectName("radioRev")
        self.gridLayout.addWidget(self.radioRev, 1, 0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.radioCram = QtWidgets.QRadioButton(parent=Dialog)
        self.radioCram.setObjectName("radioCram")
        self.gridLayout.addWidget(self.radioCram, 5, 0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.radioPreview = QtWidgets.QRadioButton(parent=Dialog)
        self.radioPreview.setObjectName("radioPreview")
        self.gridLayout.addWidget(self.radioPreview, 4, 0, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.verticalLayout.addLayout(self.gridLayout)
        self.groupBox = QtWidgets.QGroupBox(parent=Dialog)
        self.groupBox.setTitle("")
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.title = QtWidgets.QLabel(parent=self.groupBox)
        self.title.setText("...")
        self.title.setObjectName("title")
        self.verticalLayout_2.addWidget(self.title)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.preSpin = QtWidgets.QLabel(parent=self.groupBox)
        self.preSpin.setText("...")
        self.preSpin.setObjectName("preSpin")
        self.horizontalLayout.addWidget(self.preSpin)
        self.spin = QtWidgets.QSpinBox(parent=self.groupBox)
        self.spin.setObjectName("spin")
        self.horizontalLayout.addWidget(self.spin)
        self.postSpin = QtWidgets.QLabel(parent=self.groupBox)
        self.postSpin.setText("...")
        self.postSpin.setObjectName("postSpin")
        self.horizontalLayout.addWidget(self.postSpin)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.cardType = QtWidgets.QListWidget(parent=self.groupBox)
        self.cardType.setObjectName("cardType")
        item = QtWidgets.QListWidgetItem()
        self.cardType.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.cardType.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.cardType.addItem(item)
        item = QtWidgets.QListWidgetItem()
        self.cardType.addItem(item)
        self.verticalLayout_2.addWidget(self.cardType)
        self.verticalLayout.addWidget(self.groupBox)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.cardType.setCurrentRow(0)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.radioNew, self.radioRev)
        Dialog.setTabOrder(self.radioRev, self.radioForgot)
        Dialog.setTabOrder(self.radioForgot, self.radioAhead)
        Dialog.setTabOrder(self.radioAhead, self.radioPreview)
        Dialog.setTabOrder(self.radioPreview, self.radioCram)
        Dialog.setTabOrder(self.radioCram, self.spin)
        Dialog.setTabOrder(self.spin, self.buttonBox)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.actions_custom_study())
        self.radioAhead.setText(tr.custom_study_review_ahead())
        self.radioForgot.setText(tr.custom_study_review_forgotten_cards())
        self.radioNew.setText(tr.custom_study_increase_todays_new_card_limit())
        self.radioRev.setText(tr.custom_study_increase_todays_review_card_limit())
        self.radioCram.setText(tr.custom_study_study_by_card_state_or_tag())
        self.radioPreview.setText(tr.custom_study_preview_new_cards())
        __sortingEnabled = self.cardType.isSortingEnabled()
        self.cardType.setSortingEnabled(False)
        item = self.cardType.item(0)
        item.setText(tr.custom_study_new_cards_only())
        item = self.cardType.item(1)
        item.setText(tr.custom_study_due_cards_only())
        item = self.cardType.item(2)
        item.setText(tr.custom_study_all_review_cards_in_random_order())
        item = self.cardType.item(3)
        item.setText(tr.custom_study_all_cards_in_random_order_dont())
        self.cardType.setSortingEnabled(__sortingEnabled)


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""
A convenience wrapper over pysqlite.

Anki's Collection class now uses dbproxy.py instead of this class,
but this class is still used by aqt's profile manager, and a number
of add-ons rely on it.
"""

from __future__ import annotations

import os
import pprint
import time
from sqlite3 import Cursor
from sqlite3 import dbapi2 as sqlite
from typing import Any

from anki._legacy import DeprecatedNamesMixin

DBError = sqlite.Error


class DB(DeprecatedNamesMixin):
    def __init__(self, path: str, timeout: int = 0) -> None:
        self._db = sqlite.connect(path, timeout=timeout)
        self._db.text_factory = self._text_factory
        self._path = path
        self.echo = os.environ.get("DBECHO")
        self.mod = False

    def __repr__(self) -> str:
        dict_ = dict(self.__dict__)
        del dict_["_db"]
        return f"{super().__repr__()} {pprint.pformat(dict_, width=300)}"

    def execute(self, sql: str, *a: Any, **ka: Any) -> Cursor:
        canonized = sql.strip().lower()
        # mark modified?
        for stmt in "insert", "update", "delete":
            if canonized.startswith(stmt):
                self.mod = True
        start_time = time.time()
        if ka:
            # execute("...where id = :id", id=5)
            res = self._db.execute(sql, ka)
        else:
            # execute("...where id = ?", 5)
            res = self._db.execute(sql, a)
        if self.echo:
            # print a, ka
            print(sql, f"{(time.time() - start_time) * 1000:0.3f}ms")
            if self.echo == "2":
                print(a, ka)
        return res

    def executemany(self, sql: str, iterable: Any) -> None:
        self.mod = True
        start_time = time.time()
        self._db.executemany(sql, iterable)
        if self.echo:
            print(sql, f"{(time.time() - start_time) * 1000:0.3f}ms")
            if self.echo == "2":
                print(iterable)

    def commit(self) -> None:
        start_time = time.time()
        self._db.commit()
        if self.echo:
            print(f"commit {(time.time() - start_time) * 1000:0.3f}ms")

    def executescript(self, sql: str) -> None:
        self.mod = True
        if self.echo:
            print(sql)
        self._db.executescript(sql)

    def rollback(self) -> None:
        self._db.rollback()

    def scalar(self, *a: Any, **kw: Any) -> Any:
        res = self.execute(*a, **kw).fetchone()
        if res:
            return res[0]
        return None

    def all(self, *a: Any, **kw: Any) -> list:
        return self.execute(*a, **kw).fetchall()

    def first(self, *a: Any, **kw: Any) -> Any:
        cursor = self.execute(*a, **kw)
        res = cursor.fetchone()
        cursor.close()
        return res

    def list(self, *a: Any, **kw: Any) -> list:
        return [x[0] for x in self.execute(*a, **kw)]

    def close(self) -> None:
        self._db.text_factory = None
        self._db.close()

    def set_progress_handler(self, *args: Any) -> None:
        self._db.set_progress_handler(*args)

    def __enter__(self) -> "DB":
        self._db.execute("begin")
        return self

    def __exit__(self, *args: Any) -> None:
        self._db.close()

    def total_changes(self) -> Any:
        return self._db.total_changes

    def interrupt(self) -> None:
        self._db.interrupt()

    def set_autocommit(self, autocommit: bool) -> None:
        if autocommit:
            self._db.isolation_level = None
        else:
            self._db.isolation_level = ""

    # strip out invalid utf-8 when reading from db
    def _text_factory(self, data: bytes) -> str:
        return str(data, errors="ignore")

    def cursor(self, factory: type[Cursor] = Cursor) -> Cursor:
        return self._db.cursor(factory)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/backend.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/backend.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import links_pb2 as anki_dot_links__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12\x61nki/backend.proto\x12\x0c\x61nki.backend\x1a\x10\x61nki/links.proto\"R\n\x0b\x42\x61\x63kendInit\x12\x17\n\x0fpreferred_langs\x18\x01 \x03(\t\x12\x1a\n\x12locale_folder_path\x18\x02 \x01(\t\x12\x0e\n\x06server\x18\x03 \x01(\x08\"F\n\x0fI18nBackendInit\x12\x17\n\x0fpreferred_langs\x18\x04 \x03(\t\x12\x1a\n\x12locale_folder_path\x18\x05 \x01(\t\"\xa8\x05\n\x0c\x42\x61\x63kendError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12-\n\x04kind\x18\x02 \x01(\x0e\x32\x1f.anki.backend.BackendError.Kind\x12@\n\thelp_page\x18\x03 \x01(\x0e\x32(.anki.links.HelpPageLinkRequest.HelpPageH\x00\x88\x01\x01\x12\x0f\n\x07\x63ontext\x18\x04 \x01(\t\x12\x11\n\tbacktrace\x18\x05 \x01(\t\"\xe3\x03\n\x04Kind\x12\x11\n\rINVALID_INPUT\x10\x00\x12\x0e\n\nUNDO_EMPTY\x10\x01\x12\x0f\n\x0bINTERRUPTED\x10\x02\x12\x12\n\x0eTEMPLATE_PARSE\x10\x03\x12\x0c\n\x08IO_ERROR\x10\x04\x12\x0c\n\x08\x44\x42_ERROR\x10\x05\x12\x11\n\rNETWORK_ERROR\x10\x06\x12\x13\n\x0fSYNC_AUTH_ERROR\x10\x07\x12\x17\n\x13SYNC_SERVER_MESSAGE\x10\x17\x12\x14\n\x10SYNC_OTHER_ERROR\x10\x08\x12\x0e\n\nJSON_ERROR\x10\t\x12\x0f\n\x0bPROTO_ERROR\x10\n\x12\x13\n\x0fNOT_FOUND_ERROR\x10\x0b\x12\n\n\x06\x45XISTS\x10\x0c\x12\x17\n\x13\x46ILTERED_DECK_ERROR\x10\r\x12\x10\n\x0cSEARCH_ERROR\x10\x0e\x12\x16\n\x12\x43USTOM_STUDY_ERROR\x10\x0f\x12\x10\n\x0cIMPORT_ERROR\x10\x10\x12\x0b\n\x07\x44\x45LETED\x10\x11\x12\x13\n\x0f\x43\x41RD_TYPE_ERROR\x10\x12\x12\x19\n\x15\x41NKIDROID_PANIC_ERROR\x10\x13\x12\x0c\n\x08OS_ERROR\x10\x14\x12\x1e\n\x1aSCHEDULER_UPGRADE_REQUIRED\x10\x15\x12\x1e\n\x1aINVALID_CERTIFICATE_FORMAT\x10\x16\x42\x0c\n\n_help_pageB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.backend_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_BACKENDINIT']._serialized_start=54
  _globals['_BACKENDINIT']._serialized_end=136
  _globals['_I18NBACKENDINIT']._serialized_start=138
  _globals['_I18NBACKENDINIT']._serialized_end=208
  _globals['_BACKENDERROR']._serialized_start=211
  _globals['_BACKENDERROR']._serialized_end=891
  _globals['_BACKENDERROR_KIND']._serialized_start=394
  _globals['_BACKENDERROR_KIND']._serialized_end=877
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import re
from collections.abc import Callable, Iterable, Sequence
from re import Match
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    import anki._backend
    from anki.collection import Collection

# DBValue is actually Union[str, int, float, None], but if defined
# that way, every call site needs to do a type check prior to using
# the return values.
ValueFromDB = Any
Row = Sequence[ValueFromDB]

ValueForDB = Union[str, int, float, None]


class DBProxy:
    # Lifecycle
    ###############

    def __init__(self, backend: anki._backend.RustBackend) -> None:
        self._backend = backend

    # Transactions
    ###############

    def transact(self, op: Callable[[], None]) -> None:
        """Run the provided operation inside a transaction.

        Please note that all backend methods automatically wrap changes in a transaction,
        so there is no need to use this when calling methods like update_cards(), unless
        you are making other changes at the same time and want to ensure they are applied
        completely or not at all.

        If the operation throws an exception, the changes will be automatically rolled
        back.
        """

        try:
            self._backend.db_begin()
            op()
            self._backend.db_commit()
        except BaseException as e:
            self._backend.db_rollback()
            raise e

    # Querying
    ################

    def _query(
        self,
        sql: str,
        *args: ValueForDB,
        first_row_only: bool = False,
        **kwargs: ValueForDB,
    ) -> list[Row]:
        sql, args2 = emulate_named_args(sql, args, kwargs)
        # fetch rows
        return self._backend.db_query(sql, args2, first_row_only)

    # Query shortcuts
    ###################

    def all(self, sql: str, *args: ValueForDB, **kwargs: ValueForDB) -> list[Row]:
        return self._query(sql, *args, first_row_only=False, **kwargs)

    def list(
        self, sql: str, *args: ValueForDB, **kwargs: ValueForDB
    ) -> list[ValueFromDB]:
        return [x[0] for x in self._query(sql, *args, first_row_only=False, **kwargs)]

    def first(self, sql: str, *args: ValueForDB, **kwargs: ValueForDB) -> Row | None:
        rows = self._query(sql, *args, first_row_only=True, **kwargs)
        if rows:
            return rows[0]
        else:
            return None

    def scalar(self, sql: str, *args: ValueForDB, **kwargs: ValueForDB) -> ValueFromDB:
        rows = self._query(sql, *args, first_row_only=True, **kwargs)
        if rows:
            return rows[0][0]
        else:
            return None

    # execute used to return a pysqlite cursor, but now is synonymous
    # with .all()
    execute = all

    # Updates
    ################

    def executemany(self, sql: str, args: Iterable[Sequence[ValueForDB]]) -> None:
        if isinstance(args, list):
            list_args = args
        else:
            list_args = list(args)
        self._backend.db_execute_many(sql, list_args)


# convert kwargs to list format
def emulate_named_args(
    sql: str, args: tuple, kwargs: dict[str, Any]
) -> tuple[str, Sequence[ValueForDB]]:
    # nothing to do?
    if not kwargs:
        return sql, args
    print("named arguments in queries will go away in the future:", sql)
    # map args to numbers
    arg_num = {}
    args2 = list(args)
    for key, val in kwargs.items():
        args2.append(val)
        number = len(args2)
        arg_num[key] = number

    # update refs
    def repl(match: Match) -> str:
        arg = match.group(1)
        return f"?{arg_num[arg]}"

    sql = re.sub(":([a-zA-Z_0-9]+)", repl, sql)
    return sql, args2



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/media.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/media.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import notetypes_pb2 as anki_dot_notetypes__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61nki/media.proto\x12\nanki.media\x1a\x12\x61nki/generic.proto\x1a\x14\x61nki/notetypes.proto\"v\n\x12\x43heckMediaResponse\x12\x0e\n\x06unused\x18\x01 \x03(\t\x12\x0f\n\x07missing\x18\x02 \x03(\t\x12\x1b\n\x13missing_media_notes\x18\x03 \x03(\x03\x12\x0e\n\x06report\x18\x04 \x01(\t\x12\x12\n\nhave_trash\x18\x05 \x01(\x08\"(\n\x16TrashMediaFilesRequest\x12\x0e\n\x06\x66names\x18\x01 \x03(\t\"9\n\x13\x41\x64\x64MediaFileRequest\x12\x14\n\x0c\x64\x65sired_name\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x32\xa7\x03\n\x0cMediaService\x12\x41\n\nCheckMedia\x12\x13.anki.generic.Empty\x1a\x1e.anki.media.CheckMediaResponse\x12\x45\n\x0c\x41\x64\x64MediaFile\x12\x1f.anki.media.AddMediaFileRequest\x1a\x14.anki.generic.String\x12J\n\x0fTrashMediaFiles\x12\".anki.media.TrashMediaFilesRequest\x1a\x13.anki.generic.Empty\x12\x36\n\nEmptyTrash\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12\x38\n\x0cRestoreTrash\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12O\n\x17\x45xtractStaticMediaFiles\x12\x1a.anki.notetypes.NotetypeId\x1a\x18.anki.generic.StringList2\x15\n\x13\x42\x61\x63kendMediaServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.media_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_CHECKMEDIARESPONSE']._serialized_start=74
  _globals['_CHECKMEDIARESPONSE']._serialized_end=192
  _globals['_TRASHMEDIAFILESREQUEST']._serialized_start=194
  _globals['_TRASHMEDIAFILESREQUEST']._serialized_end=234
  _globals['_ADDMEDIAFILEREQUEST']._serialized_start=236
  _globals['_ADDMEDIAFILEREQUEST']._serialized_end=293
  _globals['_MEDIASERVICE']._serialized_start=296
  _globals['_MEDIASERVICE']._serialized_end=719
  _globals['_BACKENDMEDIASERVICE']._serialized_start=721
  _globals['_BACKENDMEDIASERVICE']._serialized_end=742
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Sequence

import aqt
import aqt.deckconf
import aqt.main
import aqt.operations
from anki.collection import OpChanges
from anki.models import ChangeNotetypeRequest, NotetypeId
from anki.notes import NoteId
from aqt.operations.notetype import change_notetype_of_notes
from aqt.qt import *
from aqt.utils import (
    add_close_shortcut,
    disable_help_button,
    restoreGeom,
    saveGeom,
    showWarning,
    tooltip,
    tr,
)
from aqt.webview import AnkiWebView, AnkiWebViewKind


class ChangeNotetypeDialog(QDialog):
    TITLE = "changeNotetype"
    silentlyClose = True

    def __init__(
        self,
        parent: QWidget,
        mw: aqt.main.AnkiQt,
        note_ids: Sequence[NoteId],
        notetype_id: NotetypeId,
    ) -> None:
        QDialog.__init__(self, parent)
        self.mw = mw
        self._note_ids = note_ids
        self._setup_ui(notetype_id)
        self.show()

    def _setup_ui(self, notetype_id: NotetypeId) -> None:
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.mw.garbage_collect_on_dialog_finish(self)
        self.setMinimumSize(400, 300)
        disable_help_button(self)
        restoreGeom(self, self.TITLE, default_size=(800, 800))
        add_close_shortcut(self)

        self.web = AnkiWebView(kind=AnkiWebViewKind.CHANGE_NOTETYPE)
        self.web.setVisible(False)
        self.web.load_sveltekit_page(f"change-notetype/{notetype_id}")
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.web)
        self.setLayout(layout)

        self.setWindowTitle(tr.browsing_change_notetype())

    def reject(self) -> None:
        self.web.cleanup()
        self.web = None  # type: ignore
        saveGeom(self, self.TITLE)
        QDialog.reject(self)

    def save(self, data: bytes) -> None:
        input = ChangeNotetypeRequest()
        input.ParseFromString(data)

        if not self.mw.confirm_schema_modification():
            return

        def on_done(op: OpChanges) -> None:
            tooltip(
                tr.browsing_notes_updated(count=len(input.note_ids)),
                parent=self.parentWidget(),
            )
            self.reject()

        input.note_ids.extend(self._note_ids)
        change_notetype_of_notes(parent=self, input=input).success(
            on_done
        ).run_in_background()


def change_notetype_dialog(parent: QWidget, note_ids: Sequence[NoteId]) -> None:
    try:
        notetype_id = aqt.mw.col.models.get_single_notetype_of_notes(note_ids)
    except Exception as e:
        showWarning(str(e), parent=parent)
        return

    ChangeNotetypeDialog(
        parent=parent, mw=aqt.mw, note_ids=note_ids, notetype_id=notetype_id
    )



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

# ruff: noqa: F401
import atexit
import logging
import os
import sys
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Union, cast

if "ANKI_FIRST_RUN" in os.environ:
    from .package import first_run_setup

    first_run_setup()

if sys.version_info[0] < 3 or sys.version_info[1] < 9:
    raise Exception("Anki requires Python 3.9+")

# ensure unicode filenames are supported
try:
    "テスト".encode(sys.getfilesystemencoding())
except UnicodeEncodeError:
    print("Anki requires a UTF-8 locale.")
    print("Please Google 'how to change locale on [your Linux distro]'")
    sys.exit(1)

# if sync server enabled, bypass the rest of the startup
if "--syncserver" in sys.argv:
    from anki.syncserver import run_sync_server
    from anki.utils import is_mac

    # does not return
    run_sync_server()

if sys.platform == "win32":
    from win32com.shell import shell

    shell.SetCurrentProcessExplicitAppUserModelID("Ankitects.Anki")

import argparse
import builtins
import cProfile
import getpass
import locale
import tempfile
import traceback
from pathlib import Path

import anki.lang
from anki._backend import RustBackend
from anki.buildinfo import version as _version
from anki.collection import Collection
from anki.consts import HELP_SITE
from anki.utils import checksum, is_gnome, is_lin, is_mac
from aqt import gui_hooks
from aqt.log import setup_logging
from aqt.qt import *
from aqt.qt import sip
from aqt.utils import TR, tr

if TYPE_CHECKING:
    import aqt.profiles

# compat aliases
anki.version = _version  # type: ignore
anki.Collection = Collection  # type: ignore

# we want to be able to print unicode debug info to console without
# fear of a traceback on systems with the console set to ASCII
try:
    sys.stdout.reconfigure(encoding="utf-8")  # type: ignore
    sys.stderr.reconfigure(encoding="utf-8")  # type: ignore
except AttributeError:
    if is_win:
        # On Windows without console; add a mock writer. The stderr
        # writer will be overwritten when ErrorHandler is initialized.
        sys.stderr = sys.stdout = open(os.devnull, "w", encoding="utf8")

appVersion = _version
appWebsite = "https://apps.ankiweb.net/"
appWebsiteDownloadSection = "https://apps.ankiweb.net/#download"
appDonate = "https://docs.ankiweb.net/contrib.html"
appShared = "https://ankiweb.net/shared/"
appUpdate = "https://ankiweb.net/update/desktop"
appHelpSite = HELP_SITE

from aqt.main import AnkiQt  # isort:skip
from aqt.profiles import ProfileManager, VideoDriver  # isort:skip

profiler: cProfile.Profile | None = None
mw: AnkiQt = None  # type: ignore # set on init

import aqt.forms

# Dialog manager
##########################################################################
# ensures only one copy of the window is open at once, and provides
# a way for dialogs to clean up asynchronously when collection closes

# to integrate a new window:
# - add it to _dialogs
# - define close behaviour, by either:
# -- setting silentlyClose=True to have it close immediately
# -- define a closeWithCallback() method
# - have the window opened via aqt.dialogs.open(<name>, self)
# - have a method reopen(*args), called if the user ask to open the window a second time. Arguments passed are the same than for original opening.

# - make preferences modal? cmd+q does wrong thing


from aqt import addcards, addons, browser, editcurrent, filtered_deck  # isort:skip
from aqt import stats, about, preferences, mediasync  # isort:skip


class DialogManager:
    _dialogs: dict[str, list] = {
        "AddCards": [addcards.AddCards, None],
        "AddonsDialog": [addons.AddonsDialog, None],
        "Browser": [browser.Browser, None],
        "EditCurrent": [editcurrent.EditCurrent, None],
        "FilteredDeckConfigDialog": [filtered_deck.FilteredDeckConfigDialog, None],
        "DeckStats": [stats.DeckStats, None],
        "NewDeckStats": [stats.NewDeckStats, None],
        "About": [about.show, None],
        "Preferences": [preferences.Preferences, None],
        "sync_log": [mediasync.MediaSyncDialog, None],
    }

    def open(self, name: str, *args: Any, **kwargs: Any) -> Any:
        (creator, instance) = self._dialogs[name]
        if instance:
            if instance.windowState() & Qt.WindowState.WindowMinimized:
                instance.setWindowState(
                    instance.windowState() & ~Qt.WindowState.WindowMinimized
                )
            instance.activateWindow()
            instance.raise_()
            if hasattr(instance, "reopen"):
                instance.reopen(*args, **kwargs)
        else:
            instance = creator(*args, **kwargs)
            self._dialogs[name][1] = instance
        gui_hooks.dialog_manager_did_open_dialog(self, name, instance)
        return instance

    def markClosed(self, name: str) -> None:
        self._dialogs[name] = [self._dialogs[name][0], None]

    def allClosed(self) -> bool:
        return not any(x[1] for x in self._dialogs.values())

    def closeAll(self, onsuccess: Callable[[], None]) -> bool | None:
        # can we close immediately?
        if self.allClosed():
            onsuccess()
            return None

        # ask all windows to close and await a reply
        for name, (creator, instance) in self._dialogs.items():
            if not instance:
                continue

            def callback() -> None:
                if self.allClosed():
                    onsuccess()
                else:
                    # still waiting for others to close
                    pass

            if getattr(instance, "silentlyClose", False):
                instance.close()
                callback()
            else:
                instance.closeWithCallback(callback)

        return True

    def register_dialog(
        self, name: str, creator: Callable | type, instance: Any | None = None
    ) -> None:
        """Allows add-ons to register a custom dialog to be managed by Anki's dialog
        manager, which ensures that only one copy of the window is open at once,
        and that the dialog cleans up asynchronously when the collection closes

        Please note that dialogs added in this manner need to define a close behavior
        by either:

            - setting `dialog.silentlyClose = True` to have it close immediately
            - define a `dialog.closeWithCallback()` method that is called when closed
              by the dialog manager

        TODO?: Implement more restrictive type check to ensure these requirements
        are met

        Arguments:
            name {str} -- Name/identifier of the dialog in question
            creator {Union[Callable, type]} -- A class or function to create new
                                               dialog instances with

        Keyword Arguments:
            instance {Optional[Any]} -- An optional existing instance of the dialog
                                        (default: {None})
        """
        self._dialogs[name] = [creator, instance]


dialogs = DialogManager()

# Language handling
##########################################################################
# Qt requires its translator to be installed before any GUI widgets are
# loaded, and we need the Qt language to match the i18n language or
# translated shortcuts will not work.

# A reference to the Qt translator needs to be held to prevent it from
# being immediately deallocated.
_qtrans: QTranslator | None = None


def setupLangAndBackend(
    pm: ProfileManager,
    app: QApplication,
    force: str | None = None,
    firstTime: bool = False,
) -> RustBackend:
    global _qtrans
    try:
        locale.setlocale(locale.LC_ALL, "")
    except Exception:
        pass

    # add _ and ngettext globals used by legacy code
    def fn__(arg) -> None:  # type: ignore
        print("".join(traceback.format_stack()[-2]))
        print("_ global will break in the future; please see anki/lang.py")
        return arg

    def fn_ngettext(a, b, c) -> None:  # type: ignore
        print("".join(traceback.format_stack()[-2]))
        print("ngettext global will break in the future; please see anki/lang.py")
        return b

    builtins.__dict__["_"] = fn__
    builtins.__dict__["ngettext"] = fn_ngettext

    # get lang and normalize into ja/zh-CN form
    if firstTime:
        lang = pm.meta["defaultLang"]
    else:
        lang = force or pm.meta["defaultLang"]
    lang = anki.lang.lang_to_disk_lang(lang)

    # set active language
    anki.lang.set_lang(lang)

    # switch direction for RTL languages
    if anki.lang.is_rtl(lang):
        app.setLayoutDirection(Qt.LayoutDirection.RightToLeft)
    else:
        app.setLayoutDirection(Qt.LayoutDirection.LeftToRight)

    # load qt translations
    _qtrans = QTranslator()

    qt_dir = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
    qt_lang = lang.replace("-", "_")
    if _qtrans.load(f"qtbase_{qt_lang}", qt_dir):
        app.installTranslator(_qtrans)

    backend = anki.lang.current_i18n
    assert backend is not None

    return backend


# App initialisation
##########################################################################


class NativeEventFilter(QAbstractNativeEventFilter):
    def nativeEventFilter(
        self, eventType: Any, message: Any
    ) -> tuple[bool, Any | None]:
        if eventType == "windows_generic_MSG":
            import ctypes.wintypes

            msg = ctypes.wintypes.MSG.from_address(int(message))
            if msg.message == 17:  # WM_QUERYENDSESSION
                assert mw is not None
                if mw.can_auto_sync():
                    mw.app._set_windows_shutdown_block_reason(tr.sync_syncing())
                    mw.progress.single_shot(100, mw.unloadProfileAndExit)
                    return (True, 0)
        return (False, 0)


class AnkiApp(QApplication):
    # Single instance support on Win32/Linux
    ##################################################

    appMsg = pyqtSignal(str)

    KEY = f"anki{checksum(getpass.getuser())}"
    TMOUT = 30000

    def __init__(self, argv: list[str]) -> None:
        QApplication.__init__(self, argv)
        self.installEventFilter(self)
        self._argv = argv
        self._native_event_filter = NativeEventFilter()
        if is_win:
            self.installNativeEventFilter(self._native_event_filter)

    def _set_windows_shutdown_block_reason(self, reason: str) -> None:
        if is_win:
            import ctypes
            from ctypes import windll, wintypes  # type: ignore

            assert mw is not None
            windll.user32.ShutdownBlockReasonCreate(
                wintypes.HWND.from_param(int(mw.effectiveWinId())),
                ctypes.c_wchar_p(reason),
            )

    def _unset_windows_shutdown_block_reason(self) -> None:
        if is_win:
            from ctypes import windll, wintypes  # type: ignore

            assert mw is not None
            windll.user32.ShutdownBlockReasonDestroy(
                wintypes.HWND.from_param(int(mw.effectiveWinId())),
            )

    def secondInstance(self) -> bool:
        # we accept only one command line argument. if it's missing, send
        # a blank screen to just raise the existing window
        opts, args = parseArgs(self._argv)
        buf = "raise"
        if args and args[0]:
            buf = os.path.abspath(args[0])
        if self.sendMsg(buf):
            print("Already running; reusing existing instance.")
            return True
        else:
            # send failed, so we're the first instance or the
            # previous instance died
            QLocalServer.removeServer(self.KEY)
            self._srv = QLocalServer(self)
            qconnect(self._srv.newConnection, self.onRecv)
            self._srv.listen(self.KEY)
            return False

    def sendMsg(self, txt: str) -> bool:
        sock = QLocalSocket(self)
        sock.connectToServer(self.KEY, QIODevice.OpenModeFlag.WriteOnly)
        if not sock.waitForConnected(self.TMOUT):
            # first instance or previous instance dead
            return False
        sock.write(txt.encode("utf8"))
        if not sock.waitForBytesWritten(self.TMOUT):
            # existing instance running but hung
            QMessageBox.warning(
                None,
                tr.qt_misc_anki_is_running(),
                tr.qt_misc_if_instance_is_not_responding(),
            )

            sys.exit(1)
        sock.disconnectFromServer()
        return True

    def onRecv(self) -> None:
        sock = self._srv.nextPendingConnection()
        if sock is None:
            return
        if not sock.waitForReadyRead(self.TMOUT):
            sys.stderr.write(sock.errorString())
            return
        path = bytes(cast(bytes, sock.readAll())).decode("utf8")
        self.appMsg.emit(path)  # type: ignore
        sock.disconnectFromServer()

    # OS X file/url handler
    ##################################################

    def event(self, evt: QEvent | None) -> bool:
        assert evt is not None

        if evt.type() == QEvent.Type.FileOpen:
            self.appMsg.emit(evt.file() or "raise")  # type: ignore
            return True
        return QApplication.event(self, evt)

    # Global cursor: pointer for Qt buttons
    ##################################################

    def eventFilter(self, src: Any, evt: QEvent | None) -> bool:
        assert evt is not None

        pointer_classes = (
            QPushButton,
            QCheckBox,
            QRadioButton,
            QMenu,
            QSlider,
            QToolButton,
            QTabBar,
        )
        if evt.type() in [QEvent.Type.Enter, QEvent.Type.HoverEnter]:
            if (isinstance(src, pointer_classes) and src.isEnabled()) or (
                isinstance(src, QComboBox) and not src.isEditable()
            ):
                self.setOverrideCursor(QCursor(Qt.CursorShape.PointingHandCursor))
            else:
                self.restoreOverrideCursor()
            return False

        elif evt.type() in [QEvent.Type.HoverLeave, QEvent.Type.Leave] or isinstance(
            evt, QCloseEvent
        ):
            self.restoreOverrideCursor()
            return False

        return False


def parseArgs(argv: list[str]) -> tuple[argparse.Namespace, list[str]]:
    "Returns (opts, args)."
    # py2app fails to strip this in some instances, then anki dies
    # as there's no such profile
    if is_mac and len(argv) > 1 and argv[1].startswith("-psn"):
        argv = [argv[0]]
    parser = argparse.ArgumentParser(description=f"Anki {appVersion}")
    parser.usage = "%(prog)s [OPTIONS] [file to import/add-on to install]"
    parser.add_argument("-b", "--base", help="path to base folder", default="")
    parser.add_argument("-p", "--profile", help="profile name to load", default="")
    parser.add_argument("-l", "--lang", help="interface language (en, de, etc)")
    parser.add_argument(
        "-v", "--version", help="print the Anki version and exit", action="store_true"
    )
    parser.add_argument(
        "--safemode", help="disable add-ons and automatic syncing", action="store_true"
    )
    parser.add_argument(
        "--syncserver",
        help="skip GUI and start a local sync server",
        action="store_true",
    )
    return parser.parse_known_args(argv[1:])


def setupGL(pm: aqt.profiles.ProfileManager) -> None:
    driver = pm.video_driver()
    # RHI errors are emitted multiple times so make sure we only handle them once
    driver_failed = False

    # work around pyqt loading wrong GL library
    if is_lin and not sys.platform.startswith("freebsd"):
        import ctypes

        ctypes.CDLL("libGL.so.1", ctypes.RTLD_GLOBAL)

    # catch opengl errors
    def msgHandler(category: Any, ctx: Any, msg: Any) -> None:
        if category == QtMsgType.QtDebugMsg:
            category = "debug"
        elif category == QtMsgType.QtInfoMsg:
            category = "info"
        elif category == QtMsgType.QtWarningMsg:
            category = "warning"
        elif category == QtMsgType.QtCriticalMsg:
            category = "critical"
        elif category == QtMsgType.QtDebugMsg:
            category = "debug"
        elif category == QtMsgType.QtFatalMsg:
            category = "fatal"
        else:
            category = "unknown"
        context = ""
        if ctx.file:
            context += f"{ctx.file}:"
        if ctx.line:
            context += f"{ctx.line},"
        if ctx.function:
            context += f"{ctx.function}"
        if context:
            context = f"'{context}'"

        nonlocal driver_failed
        if not driver_failed and (
            "Failed to create OpenGL context" in msg
            # Based on the message Qt6 shows to the user; have not tested whether
            # we can actually capture this or not.
            or "Failed to initialize graphics backend" in msg
            # RHI backend
            or "Failed to create QRhi" in msg
            or "Failed to get a QRhi" in msg
        ):
            QMessageBox.critical(
                None,
                tr.qt_misc_error(),
                tr.qt_misc_error_loading_graphics_driver(
                    mode=driver.value,
                    context=context,
                ),
            )
            pm.set_video_driver(driver.next())
            driver_failed = True
            return
        else:
            print(f"Qt {category}: {msg} {context}")

    qInstallMessageHandler(msgHandler)
    atexit.register(qInstallMessageHandler, None)

    if driver == VideoDriver.OpenGL:
        # Leaving QT_OPENGL unset appears to sometimes produce different results
        # to explicitly setting it to 'auto'; the former seems to be more compatible.
        if qtmajor > 5:
            QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.OpenGL)
    elif driver in (VideoDriver.Software, VideoDriver.ANGLE):
        if is_win:
            # on Windows, this appears to be sufficient
            # On Qt6, ANGLE is excluded by the enum.
            os.environ["QT_OPENGL"] = driver.value
        elif is_mac:
            QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_UseSoftwareOpenGL)
        elif is_lin:
            if "QTWEBENGINE_CHROMIUM_FLAGS" not in os.environ:
                os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--disable-gpu"
        if qtmajor > 5:
            QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.Software)
    elif driver == VideoDriver.Metal:
        QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.Metal)
    elif driver == VideoDriver.Vulkan:
        QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.Vulkan)
    elif driver == VideoDriver.Direct3D:
        QQuickWindow.setGraphicsApi(QSGRendererInterface.GraphicsApi.Direct3D11)


PROFILE_CODE = os.environ.get("ANKI_PROFILE_CODE")


def write_profile_results() -> None:
    assert profiler is not None

    profiler.disable()
    profile = "out/anki.prof"
    profiler.dump_stats(profile)


def run() -> None:
    print(f"Starting Anki {_version}...")
    try:
        _run()
    except Exception:
        traceback.print_exc()
        QMessageBox.critical(
            None,
            "Startup Error",
            f"Please notify support of this error:\n\n{traceback.format_exc()}",
        )


def _run(argv: list[str] | None = None, exec: bool = True) -> AnkiApp | None:
    """Start AnkiQt application or reuse an existing instance if one exists.

    If the function is invoked with exec=False, the AnkiQt will not enter
    the main event loop - instead the application object will be returned.

    The 'exec' and 'argv' arguments will be useful for testing purposes.

    If no 'argv' is supplied then 'sys.argv' will be used.
    """
    global mw
    global profiler

    if argv is None:
        argv = sys.argv

    # parse args
    opts, args = parseArgs(argv)

    if opts.version:
        print(f"Anki {appVersion}")
        return None

    if PROFILE_CODE:
        profiler = cProfile.Profile()
        profiler.enable()

    x11_available = os.getenv("DISPLAY")
    wayland_configured = qtmajor > 5 and (
        os.getenv("QT_QPA_PLATFORM") == "wayland" or os.getenv("WAYLAND_DISPLAY")
    )
    wayland_forced = os.getenv("ANKI_WAYLAND")

    if is_gnome and wayland_configured:
        if wayland_forced or not x11_available:
            # Work around broken fractional scaling in Wayland
            # https://bugreports.qt.io/browse/QTBUG-113574
            os.environ["QT_SCALE_FACTOR_ROUNDING_POLICY"] = "RoundPreferFloor"
            if not x11_available:
                print(
                    "Trying to use X11, but it is not available. Falling back to Wayland, which has some bugs:"
                )
                print("https://github.com/ankitects/anki/issues/1767")
        else:
            # users need to opt in to wayland support, given the issues it has
            print("Wayland support is disabled by default due to bugs:")
            print("https://github.com/ankitects/anki/issues/1767")
            print("You can force it on with an env var: ANKI_WAYLAND=1")
            os.environ["QT_QPA_PLATFORM"] = "xcb"

    # profile manager
    i18n_setup = False
    pm = None
    try:
        base_folder = ProfileManager.get_created_base_folder(opts.base)

        # default to specified/system language before getting user's preference so that we can localize some more strings
        lang = anki.lang.get_def_lang(opts.lang)
        anki.lang.set_lang(lang[1])
        i18n_setup = True

        pm = ProfileManager(base_folder)
        pmLoadResult = pm.setupMeta()

        Collection.initialize_backend_logging()
    except Exception:
        # will handle below
        traceback.print_exc()
        pm = None

    if pm:
        # gl workarounds
        setupGL(pm)
        # apply user-provided scale factor
        os.environ["QT_SCALE_FACTOR"] = str(pm.uiScale())

    # Opt-in to full HiDPI support?
    if not os.environ.get("ANKI_NOHIGHDPI") and qtmajor == 5:
        QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)  # type: ignore
        QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)  # type: ignore
        os.environ["QT_ENABLE_HIGHDPI_SCALING"] = "1"
        os.environ["QT_SCALE_FACTOR_ROUNDING_POLICY"] = "PassThrough"

    # Opt-in to software rendering?
    if os.environ.get("ANKI_SOFTWAREOPENGL"):
        QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_UseSoftwareOpenGL)

    # Fix an issue on Windows, where Ctrl+Alt shortcuts are triggered by AltGr,
    # preventing users from typing things like "@" through AltGr+Q on a German
    # keyboard.
    if is_win and "QT_QPA_PLATFORM" not in os.environ:
        os.environ["QT_QPA_PLATFORM"] = "windows:altgr"

    # create the app
    QCoreApplication.setApplicationName("Anki")
    QGuiApplication.setDesktopFileName("anki")
    app = AnkiApp(argv)
    if app.secondInstance():
        # we've signaled the primary instance, so we should close
        return None

    if not pm:
        if i18n_setup:
            QMessageBox.critical(
                None,
                tr.qt_misc_error(),
                tr.profiles_could_not_create_data_folder(),
            )
        else:
            QMessageBox.critical(None, "Startup Failed", "Unable to create data folder")
        return None

    setup_logging(
        pm.addon_logs(),
        level=logging.DEBUG if int(os.getenv("ANKIDEV", "0")) else logging.INFO,
    )

    # disable icons on mac; this must be done before window created
    if is_mac:
        app.setAttribute(Qt.ApplicationAttribute.AA_DontShowIconsInMenus)

    # disable help button in title bar on qt versions that support it
    if is_win and qtmajor == 5 and qtminor >= 10:
        QApplication.setAttribute(Qt.AA_DisableWindowContextHelpButton)  # type: ignore

    # proxy configured?
    from urllib.request import getproxies, proxy_bypass

    disable_proxies = False
    try:
        if "http" in getproxies():
            # if it's not set up to bypass localhost, we'll
            # need to disable proxies in the webviews
            if not proxy_bypass("127.0.0.1"):
                disable_proxies = True
    except UnicodeDecodeError:
        # proxy_bypass can't handle unicode in hostnames; assume we need
        # to disable proxies
        disable_proxies = True

    if disable_proxies:
        print("webview proxy use disabled")
        proxy = QNetworkProxy()
        proxy.setType(QNetworkProxy.ProxyType.NoProxy)
        QNetworkProxy.setApplicationProxy(proxy)

    # we must have a usable temp dir
    try:
        tempfile.gettempdir()
    except Exception:
        QMessageBox.critical(
            None,
            tr.qt_misc_error(),
            tr.qt_misc_no_temp_folder(),
        )
        return None

    # make image resources available
    from aqt.utils import aqt_data_folder

    QDir.addSearchPath("icons", os.path.join(aqt_data_folder(), "qt", "icons"))

    if pmLoadResult.firstTime:
        pm.setDefaultLang(lang[0])

    if pmLoadResult.loadError:
        QMessageBox.warning(
            None,
            tr.profiles_prefs_corrupt_title(),
            tr.profiles_prefs_file_is_corrupt(),
        )

    if opts.profile:
        pm.openProfile(opts.profile)

    # i18n & backend
    backend = setupLangAndBackend(pm, app, opts.lang, pmLoadResult.firstTime)

    driver = pm.video_driver()
    if is_lin and driver == VideoDriver.OpenGL:
        from aqt.utils import gfxDriverIsBroken

        if gfxDriverIsBroken():
            pm.set_video_driver(driver.next())
            QMessageBox.critical(
                None,
                tr.qt_misc_error(),
                tr.qt_misc_incompatible_video_driver(),
            )
            sys.exit(1)

    # load the main window
    import aqt.main

    mw = aqt.main.AnkiQt(app, pm, backend, opts, args)
    if exec:
        print("Starting main loop...")
        app.exec()
    else:
        return app

    if PROFILE_CODE:
        write_profile_results()

    return None



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import json
from dataclasses import dataclass
from urllib.parse import quote

import aqt
import aqt.deckconf
import aqt.main
import aqt.operations
from aqt.qt import *
from aqt.utils import add_close_shortcut, disable_help_button, restoreGeom, saveGeom, tr
from aqt.webview import AnkiWebView, AnkiWebViewKind


@dataclass
class ImportArgs:
    path: str
    title = "importLog"
    kind = AnkiWebViewKind.IMPORT_LOG
    ts_page = "import-page"

    def args_json(self) -> str:
        return json.dumps(self.path)


class JsonFileArgs(ImportArgs):
    def args_json(self) -> str:
        return json.dumps(dict(type="json_file", path=self.path))


class CsvArgs(ImportArgs):
    title = "csv import"
    kind = AnkiWebViewKind.IMPORT_CSV
    ts_page = "import-csv"


class AnkiPackageArgs(ImportArgs):
    title = "anki package import"
    kind = AnkiWebViewKind.IMPORT_ANKI_PACKAGE
    ts_page = "import-anki-package"


class ImportDialog(QDialog):
    DEFAULT_SIZE = (800, 600)
    MIN_SIZE = (400, 300)
    silentlyClose = True

    def __init__(self, mw: aqt.main.AnkiQt, args: ImportArgs) -> None:
        QDialog.__init__(self, mw, Qt.WindowType.Window)
        self.mw = mw
        self.args = args
        self._setup_ui()
        self.show()

    def _setup_ui(self) -> None:
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.mw.garbage_collect_on_dialog_finish(self)
        self.setMinimumSize(*self.MIN_SIZE)
        disable_help_button(self)
        restoreGeom(self, self.args.title, default_size=self.DEFAULT_SIZE)
        add_close_shortcut(self)

        self.web: AnkiWebView | None = AnkiWebView(kind=self.args.kind)
        self.web.setVisible(False)
        self.web.load_sveltekit_page(f"{self.args.ts_page}/{quote(self.args.path)}")
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.web)
        self.setLayout(layout)
        restoreGeom(self, self.args.title, default_size=(800, 800))

        self.setWindowTitle(tr.decks_import_file())

    def reject(self) -> None:
        if self.mw.col and self.windowModality() == Qt.WindowModality.ApplicationModal:
            self.mw.col.set_wants_abort()
        assert self.web is not None
        self.web.cleanup()
        self.web = None
        saveGeom(self, self.args.title)
        QDialog.reject(self)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

from abc import ABC, abstractmethod, abstractproperty
from collections.abc import Sequence
from typing import cast

from anki.browser import BrowserConfig
from anki.cards import Card, CardId
from anki.collection import Collection
from anki.errors import NotFoundError
from anki.notes import Note, NoteId
from anki.utils import ids2str
from aqt.browser.table import Column, ItemId, ItemList


class ItemState(ABC):
    GEOMETRY_KEY_PREFIX: str
    SORT_COLUMN_KEY: str
    SORT_BACKWARDS_KEY: str
    _active_columns: list[str]

    def __init__(self, col: Collection) -> None:
        self.col = col
        self._sort_column = self.col.get_config(self.SORT_COLUMN_KEY)
        self._sort_backwards = self.col.get_config(self.SORT_BACKWARDS_KEY, False)

    def is_notes_mode(self) -> bool:
        """Return True if the state is a NoteState."""
        return isinstance(self, NoteState)

    # Stateless Helpers

    def note_ids_from_card_ids(self, items: Sequence[ItemId]) -> Sequence[NoteId]:
        assert self.col.db is not None
        return self.col.db.list(
            f"select distinct nid from cards where id in {ids2str(items)}"
        )

    def card_ids_from_note_ids(self, items: Sequence[ItemId]) -> Sequence[CardId]:
        assert self.col.db is not None
        return self.col.db.list(f"select id from cards where nid in {ids2str(items)}")

    def column_key_at(self, index: int) -> str:
        return self._active_columns[index]

    def column_label(self, column: Column) -> str:
        return (
            column.notes_mode_label if self.is_notes_mode() else column.cards_mode_label
        )

    def column_tooltip(self, column: Column) -> str:
        if self.is_notes_mode():
            return column.notes_mode_tooltip
        return column.cards_mode_tooltip

    # Columns and sorting

    # abstractproperty is deprecated but used due to mypy limitations
    # (https://github.com/python/mypy/issues/1362)
    @abstractproperty
    def active_columns(self) -> list[str]:
        """Return the saved or default columns for the state."""

    @abstractmethod
    def toggle_active_column(self, column: str) -> None:
        """Add or remove an active column."""

    @property
    def sort_column(self) -> str:
        return self._sort_column

    @sort_column.setter
    def sort_column(self, column: str) -> None:
        self.col.set_config(self.SORT_COLUMN_KEY, column)
        self._sort_column = column

    @property
    def sort_backwards(self) -> bool:
        "If true, descending order."
        return self._sort_backwards

    @sort_backwards.setter
    def sort_backwards(self, order: bool) -> None:
        self.col.set_config(self.SORT_BACKWARDS_KEY, order)
        self._sort_backwards = order

    # Get objects

    @abstractmethod
    def get_card(self, item: ItemId) -> Card:
        """Return the item if it's a card or its first card if it's a note."""

    @abstractmethod
    def get_note(self, item: ItemId) -> Note:
        """Return the item if it's a note or its note if it's a card."""

    # Get ids

    @abstractmethod
    def find_items(
        self, search: str, order: bool | str | Column, reverse: bool
    ) -> Sequence[ItemId]:
        """Return the item ids fitting the given search and order."""

    @abstractmethod
    def get_item_from_card_id(self, card: CardId) -> ItemId:
        """Return the appropriate item id for a card id."""

    @abstractmethod
    def get_card_ids(self, items: Sequence[ItemId]) -> Sequence[CardId]:
        """Return the card ids for the given item ids."""

    @abstractmethod
    def get_note_ids(self, items: Sequence[ItemId]) -> Sequence[NoteId]:
        """Return the note ids for the given item ids."""

    # Toggle

    @abstractmethod
    def toggle_state(self) -> ItemState:
        """Return an instance of the other state."""

    @abstractmethod
    def get_new_items(self, old_items: Sequence[ItemId]) -> ItemList:
        """Given a list of ids from the other state, return the corresponding ids for this state."""


class CardState(ItemState):
    GEOMETRY_KEY_PREFIX = "editor"
    SORT_COLUMN_KEY = BrowserConfig.CARDS_SORT_COLUMN_KEY
    SORT_BACKWARDS_KEY = BrowserConfig.CARDS_SORT_BACKWARDS_KEY

    def __init__(self, col: Collection) -> None:
        super().__init__(col)
        self._active_columns = self.col.load_browser_card_columns()

    @property
    def active_columns(self) -> list[str]:
        return self._active_columns

    def toggle_active_column(self, column: str) -> None:
        if column in self._active_columns:
            self._active_columns.remove(column)
        else:
            self._active_columns.append(column)
        self.col.set_browser_card_columns(self._active_columns)

    def get_card(self, item: ItemId) -> Card:
        return self.col.get_card(CardId(item))

    def get_note(self, item: ItemId) -> Note:
        return self.get_card(item).note()

    def find_items(
        self, search: str, order: bool | str | Column, reverse: bool
    ) -> Sequence[ItemId]:
        return self.col.find_cards(search, order, reverse)

    def get_item_from_card_id(self, card: CardId) -> ItemId:
        return card

    def get_card_ids(self, items: Sequence[ItemId]) -> Sequence[CardId]:
        return cast(Sequence[CardId], items)

    def get_note_ids(self, items: Sequence[ItemId]) -> Sequence[NoteId]:
        return super().note_ids_from_card_ids(items)

    def toggle_state(self) -> NoteState:
        return NoteState(self.col)

    def get_new_items(self, old_items: Sequence[ItemId]) -> Sequence[CardId]:
        return super().card_ids_from_note_ids(old_items)


class NoteState(ItemState):
    GEOMETRY_KEY_PREFIX = "editorNotesMode"
    SORT_COLUMN_KEY = BrowserConfig.NOTES_SORT_COLUMN_KEY
    SORT_BACKWARDS_KEY = BrowserConfig.NOTES_SORT_BACKWARDS_KEY

    def __init__(self, col: Collection) -> None:
        super().__init__(col)
        self._active_columns = self.col.load_browser_note_columns()

    @property
    def active_columns(self) -> list[str]:
        return self._active_columns

    def toggle_active_column(self, column: str) -> None:
        if column in self._active_columns:
            self._active_columns.remove(column)
        else:
            self._active_columns.append(column)
        self.col.set_browser_note_columns(self._active_columns)

    def get_card(self, item: ItemId) -> Card:
        if cards := self.get_note(item).cards():
            return cards[0]
        raise NotFoundError("card not found", None, None, None)

    def get_note(self, item: ItemId) -> Note:
        return self.col.get_note(NoteId(item))

    def find_items(
        self, search: str, order: bool | str | Column, reverse: bool
    ) -> Sequence[ItemId]:
        return self.col.find_notes(search, order, reverse)

    def get_item_from_card_id(self, card: CardId) -> ItemId:
        return self.col.get_card(card).note().id

    def get_card_ids(self, items: Sequence[ItemId]) -> Sequence[CardId]:
        return super().card_ids_from_note_ids(items)

    def get_note_ids(self, items: Sequence[ItemId]) -> Sequence[NoteId]:
        return cast(Sequence[NoteId], items)

    def toggle_state(self) -> CardState:
        return CardState(self.col)

    def get_new_items(self, old_items: Sequence[ItemId]) -> Sequence[NoteId]:
        return super().note_ids_from_card_ids(old_items)



# ruff: noqa: F401
from . import (
    about,
    addcards,
    addfield,
    addmodel,
    addonconf,
    addons,
    browser,
    browserdisp,
    browseropts,
    changemap,
    changemodel,
    clayout_top,
    customstudy,
    dconf,
    debug,
    editcurrent,
    edithtml,
    emptycards,
    exporting,
    fields,
    filtered_deck,
    finddupes,
    findreplace,
    forget,
    getaddons,
    importing,
    main,
    modelopts,
    models,
    preferences,
    preview,
    profiles,
    progress,
    reposition,
    setgroup,
    setlang,
    stats,
    studydeck,
    synclog,
    taglimit,
    template,
    widgets,
)



# Form implementation generated from reading ui file 'qt/aqt/forms/stats.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(607, 556)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.web = StatsWebView(parent=Dialog)
        self.web.setProperty("url", QtCore.QUrl("about:blank"))
        self.web.setObjectName("web")
        self.verticalLayout.addWidget(self.web)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setContentsMargins(16, 6, 16, 6)
        self.horizontalLayout_3.setSpacing(8)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.groupBox_2 = QtWidgets.QGroupBox(parent=Dialog)
        self.groupBox_2.setTitle("")
        self.groupBox_2.setObjectName("groupBox_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.groupBox_2)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.groups = QtWidgets.QRadioButton(parent=self.groupBox_2)
        self.groups.setText("deck")
        self.groups.setChecked(True)
        self.groups.setObjectName("groups")
        self.horizontalLayout_2.addWidget(self.groups, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.all = QtWidgets.QRadioButton(parent=self.groupBox_2)
        self.all.setText("collection")
        self.all.setObjectName("all")
        self.horizontalLayout_2.addWidget(self.all, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.horizontalLayout_3.addWidget(self.groupBox_2)
        self.groupBox = QtWidgets.QGroupBox(parent=Dialog)
        self.groupBox.setTitle("")
        self.groupBox.setObjectName("groupBox")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.groupBox)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.month = QtWidgets.QRadioButton(parent=self.groupBox)
        self.month.setText("1 month")
        self.month.setChecked(True)
        self.month.setObjectName("month")
        self.horizontalLayout.addWidget(self.month, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.year = QtWidgets.QRadioButton(parent=self.groupBox)
        self.year.setText("1 year")
        self.year.setObjectName("year")
        self.horizontalLayout.addWidget(self.year, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.life = QtWidgets.QRadioButton(parent=self.groupBox)
        self.life.setText("deck life")
        self.life.setObjectName("life")
        self.horizontalLayout.addWidget(self.life, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.horizontalLayout_3.addWidget(self.groupBox)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setContentsMargins(-1, 4, -1, 0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.deckArea = QtWidgets.QWidget(parent=Dialog)
        self.deckArea.setObjectName("deckArea")
        self.horizontalLayout_4.addWidget(self.deckArea, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.horizontalLayout_3.addLayout(self.horizontalLayout_4)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Close)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.horizontalLayout_3.addWidget(self.buttonBox, 0, QtCore.Qt.AlignmentFlag.AlignRight)
        self.verticalLayout.addLayout(self.horizontalLayout_3)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.statistics_title())
from aqt.webview import StatsWebView


# Form implementation generated from reading ui file 'qt/aqt/forms/fields.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(567, 438)
        Dialog.setModal(True)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.fieldList = QtWidgets.QListWidget(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fieldList.sizePolicy().hasHeightForWidth())
        self.fieldList.setSizePolicy(sizePolicy)
        self.fieldList.setMinimumSize(QtCore.QSize(50, 60))
        self.fieldList.setObjectName("fieldList")
        self.horizontalLayout.addWidget(self.fieldList)
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.fieldAdd = QtWidgets.QPushButton(parent=Dialog)
        self.fieldAdd.setObjectName("fieldAdd")
        self.verticalLayout_3.addWidget(self.fieldAdd)
        self.fieldDelete = QtWidgets.QPushButton(parent=Dialog)
        self.fieldDelete.setObjectName("fieldDelete")
        self.verticalLayout_3.addWidget(self.fieldDelete)
        self.fieldRename = QtWidgets.QPushButton(parent=Dialog)
        self.fieldRename.setObjectName("fieldRename")
        self.verticalLayout_3.addWidget(self.fieldRename)
        self.fieldPosition = QtWidgets.QPushButton(parent=Dialog)
        self.fieldPosition.setObjectName("fieldPosition")
        self.verticalLayout_3.addWidget(self.fieldPosition)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_3.addItem(spacerItem)
        self.horizontalLayout.addLayout(self.verticalLayout_3)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self._2 = QtWidgets.QGridLayout()
        self._2.setObjectName("_2")
        self.label_description = QtWidgets.QLabel(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_description.sizePolicy().hasHeightForWidth())
        self.label_description.setSizePolicy(sizePolicy)
        self.label_description.setObjectName("label_description")
        self._2.addWidget(self.label_description, 0, 0, 1, 1)
        self.label_font = QtWidgets.QLabel(parent=Dialog)
        self.label_font.setObjectName("label_font")
        self._2.addWidget(self.label_font, 1, 0, 1, 1)
        self.fontSize = QtWidgets.QSpinBox(parent=Dialog)
        self.fontSize.setMinimum(5)
        self.fontSize.setMaximum(300)
        self.fontSize.setObjectName("fontSize")
        self._2.addWidget(self.fontSize, 1, 2, 1, 1)
        self.label_sort = QtWidgets.QLabel(parent=Dialog)
        self.label_sort.setObjectName("label_sort")
        self._2.addWidget(self.label_sort, 2, 0, 1, 1)
        self.fontFamily = QtWidgets.QFontComboBox(parent=Dialog)
        self.fontFamily.setMinimumSize(QtCore.QSize(0, 25))
        self.fontFamily.setObjectName("fontFamily")
        self._2.addWidget(self.fontFamily, 1, 1, 1, 1)
        self.rtl = QtWidgets.QCheckBox(parent=Dialog)
        self.rtl.setObjectName("rtl")
        self._2.addWidget(self.rtl, 3, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.plainTextByDefault = QtWidgets.QCheckBox(parent=Dialog)
        self.plainTextByDefault.setEnabled(True)
        self.plainTextByDefault.setObjectName("plainTextByDefault")
        self._2.addWidget(self.plainTextByDefault, 4, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.fieldDescription = QtWidgets.QLineEdit(parent=Dialog)
        self.fieldDescription.setObjectName("fieldDescription")
        self._2.addWidget(self.fieldDescription, 0, 1, 1, 2)
        self.sortField = QtWidgets.QRadioButton(parent=Dialog)
        self.sortField.setObjectName("sortField")
        self._2.addWidget(self.sortField, 2, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.collapseByDefault = QtWidgets.QCheckBox(parent=Dialog)
        self.collapseByDefault.setEnabled(True)
        self.collapseByDefault.setObjectName("collapseByDefault")
        self._2.addWidget(self.collapseByDefault, 5, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.excludeFromSearch = QtWidgets.QCheckBox(parent=Dialog)
        self.excludeFromSearch.setEnabled(True)
        self.excludeFromSearch.setObjectName("excludeFromSearch")
        self._2.addWidget(self.excludeFromSearch, 6, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.verticalLayout.addLayout(self._2)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Help|QtWidgets.QDialogButtonBox.StandardButton.Save)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.fieldList, self.fieldAdd)
        Dialog.setTabOrder(self.fieldAdd, self.fieldDelete)
        Dialog.setTabOrder(self.fieldDelete, self.fieldRename)
        Dialog.setTabOrder(self.fieldRename, self.fieldPosition)
        Dialog.setTabOrder(self.fieldPosition, self.fieldDescription)
        Dialog.setTabOrder(self.fieldDescription, self.fontFamily)
        Dialog.setTabOrder(self.fontFamily, self.fontSize)
        Dialog.setTabOrder(self.fontSize, self.sortField)
        Dialog.setTabOrder(self.sortField, self.rtl)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.editing_fields())
        self.fieldAdd.setText(tr.actions_add())
        self.fieldDelete.setText(tr.actions_delete())
        self.fieldRename.setText(tr.actions_rename())
        self.fieldPosition.setText(tr.actions_reposition())
        self.label_description.setText(tr.fields_description())
        self.label_font.setText(tr.fields_editing_font())
        self.label_sort.setText(tr.actions_options())
        self.rtl.setText(tr.fields_reverse_text_direction_rtl())
        self.plainTextByDefault.setText(tr.fields_html_by_default())
        self.fieldDescription.setPlaceholderText(tr.fields_description_placeholder())
        self.sortField.setText(tr.fields_sort_by_this_field_in_the())
        self.collapseByDefault.setText(tr.fields_collapse_by_default())
        self.excludeFromSearch.setText(tr.fields_exclude_from_search())


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import html
from copy import deepcopy
from dataclasses import dataclass
from typing import Any

import aqt
import aqt.operations
from anki.collection import Collection, OpChanges
from anki.decks import DeckCollapseScope, DeckId, DeckTreeNode
from aqt import AnkiQt, gui_hooks
from aqt.deckoptions import display_options_for_deck_id
from aqt.operations import QueryOp
from aqt.operations.deck import (
    add_deck_dialog,
    remove_decks,
    rename_deck,
    reparent_decks,
    set_current_deck,
    set_deck_collapsed,
)
from aqt.qt import *
from aqt.sound import av_player
from aqt.toolbar import BottomBar
from aqt.utils import getOnlyText, openLink, shortcut, showInfo, tr


class DeckBrowserBottomBar:
    def __init__(self, deck_browser: DeckBrowser) -> None:
        self.deck_browser = deck_browser


@dataclass
class RenderData:
    """Data from collection that is required to show the page."""

    tree: DeckTreeNode
    current_deck_id: DeckId
    studied_today: str
    sched_upgrade_required: bool


@dataclass
class DeckBrowserContent:
    """Stores sections of HTML content that the deck browser will be
    populated with.

    Attributes:
        tree {str} -- HTML of the deck tree section
        stats {str} -- HTML of the stats section
    """

    tree: str
    stats: str


@dataclass
class RenderDeckNodeContext:
    current_deck_id: DeckId


class DeckBrowser:
    _render_data: RenderData

    def __init__(self, mw: AnkiQt) -> None:
        self.mw = mw
        self.web = mw.web
        self.bottom = BottomBar(mw, mw.bottomWeb)
        self.scrollPos = QPoint(0, 0)
        self._refresh_needed = False

    def show(self) -> None:
        av_player.stop_and_clear_queue()
        self.web.set_bridge_command(self._linkHandler, self)
        # redraw top bar for theme change
        self.mw.toolbar.redraw()
        self.refresh()

    def refresh(self) -> None:
        self._renderPage()
        self._refresh_needed = False

    def refresh_if_needed(self) -> None:
        if self._refresh_needed:
            self.refresh()

    def op_executed(
        self, changes: OpChanges, handler: object | None, focused: bool
    ) -> bool:
        if changes.study_queues and handler is not self:
            self._refresh_needed = True

        if focused:
            self.refresh_if_needed()

        return self._refresh_needed

    # Event handlers
    ##########################################################################

    def _linkHandler(self, url: str) -> Any:
        if ":" in url:
            (cmd, arg) = url.split(":", 1)
        else:
            cmd = url
            arg = ""
        if cmd == "open":
            self.set_current_deck(DeckId(int(arg)))
        elif cmd == "opts":
            self._showOptions(arg)
        elif cmd == "shared":
            self._onShared()
        elif cmd == "import":
            self.mw.onImport()
        elif cmd == "create":
            self._on_create()
        elif cmd == "drag":
            source, target = arg.split(",")
            self._handle_drag_and_drop(DeckId(int(source)), DeckId(int(target or 0)))
        elif cmd == "collapse":
            self._collapse(DeckId(int(arg)))
        elif cmd == "v2upgrade":
            self._confirm_upgrade()
        elif cmd == "v2upgradeinfo":
            if self.mw.col.sched_ver() == 1:
                openLink("https://faqs.ankiweb.net/the-anki-2.1-scheduler.html")
            else:
                openLink("https://faqs.ankiweb.net/the-2021-scheduler.html")
        elif cmd == "select":
            set_current_deck(
                parent=self.mw, deck_id=DeckId(int(arg))
            ).run_in_background()
        return False

    def set_current_deck(self, deck_id: DeckId) -> None:
        set_current_deck(parent=self.mw, deck_id=deck_id).success(
            lambda _: self.mw.onOverview()
        ).run_in_background(initiator=self)

    # HTML generation
    ##########################################################################

    _body = """
<center>
<table cellspacing=0 cellpadding=3>
%(tree)s
</table>

<br>
%(stats)s
</center>
"""

    def _renderPage(self, reuse: bool = False) -> None:
        if not reuse:

            def get_data(col: Collection) -> RenderData:
                return RenderData(
                    tree=col.sched.deck_due_tree(),
                    current_deck_id=col.decks.get_current_id(),
                    studied_today=col.studied_today(),
                    sched_upgrade_required=not col.v3_scheduler(),
                )

            def success(output: RenderData) -> None:
                self._render_data = output
                self.__renderPage(None)

            QueryOp(
                parent=self.mw,
                op=get_data,
                success=success,
            ).run_in_background()
        else:
            self.web.evalWithCallback("window.pageYOffset", self.__renderPage)

    def __renderPage(self, offset: int | None) -> None:
        data = self._render_data
        content = DeckBrowserContent(
            tree=self._renderDeckTree(data.tree),
            stats=self._renderStats(),
        )
        gui_hooks.deck_browser_will_render_content(self, content)
        self.web.stdHtml(
            self._v1_upgrade_message(data.sched_upgrade_required)
            + self._body % content.__dict__,
            css=["css/deckbrowser.css"],
            js=[
                "js/vendor/jquery.min.js",
                "js/vendor/jquery-ui.min.js",
                "js/deckbrowser.js",
            ],
            context=self,
        )
        self._drawButtons()
        if offset is not None:
            self._scrollToOffset(offset)
        gui_hooks.deck_browser_did_render(self)

    def _scrollToOffset(self, offset: int) -> None:
        self.web.eval("window.scrollTo(0, %d, 'instant');" % offset)

    def _renderStats(self) -> str:
        return '<div id="studiedToday"><span>{}</span></div>'.format(
            self._render_data.studied_today
        )

    def _renderDeckTree(self, top: DeckTreeNode) -> str:
        buf = """
<tr><th colspan=5 align=start>{}</th>
<th class=count>{}</th>
<th class=count>{}</th>
<th class=count>{}</th>
<th class=optscol></th></tr>""".format(
            tr.decks_deck(),
            tr.actions_new(),
            tr.decks_learn_header(),
            tr.decks_review_header(),
        )
        buf += self._topLevelDragRow()

        ctx = RenderDeckNodeContext(current_deck_id=self._render_data.current_deck_id)

        for child in top.children:
            buf += self._render_deck_node(child, ctx)

        return buf

    def _render_deck_node(self, node: DeckTreeNode, ctx: RenderDeckNodeContext) -> str:
        if node.collapsed:
            prefix = "+"
        else:
            prefix = "-"

        def indent() -> str:
            return "&nbsp;" * 6 * (node.level - 1)

        if node.deck_id == ctx.current_deck_id:
            klass = "deck current"
        else:
            klass = "deck"

        buf = (
            "<tr class='%s' id='%d' onclick='if(event.shiftKey) return pycmd(\"select:%d\")'>"
            % (
                klass,
                node.deck_id,
                node.deck_id,
            )
        )
        # deck link
        if node.children:
            collapse = (
                "<a class=collapse href=# onclick='return pycmd(\"collapse:%d\")'>%s</a>"
                % (node.deck_id, prefix)
            )
        else:
            collapse = "<span class=collapse></span>"
        if node.filtered:
            extraclass = "filtered"
        else:
            extraclass = ""
        buf += """

        <td class=decktd colspan=5>%s%s<a class="deck %s"
        href=# onclick="return pycmd('open:%d')">%s</a></td>""" % (
            indent(),
            collapse,
            extraclass,
            node.deck_id,
            html.escape(node.name),
        )

        # due counts
        def nonzeroColour(cnt: int, klass: str) -> str:
            if not cnt:
                klass = "zero-count"
            return f'<span class="{klass}">{cnt}</span>'

        review = nonzeroColour(node.review_count, "review-count")
        learn = nonzeroColour(node.learn_count, "learn-count")

        buf += ("<td align=end>%s</td>" * 3) % (
            nonzeroColour(node.new_count, "new-count"),
            learn,
            review,
        )
        # options
        buf += (
            "<td align=center class=opts><a onclick='return pycmd(\"opts:%d\");'>"
            "<img src='/_anki/imgs/gears.svg' class=gears></a></td></tr>" % node.deck_id
        )
        # children
        if not node.collapsed:
            for child in node.children:
                buf += self._render_deck_node(child, ctx)
        return buf

    def _topLevelDragRow(self) -> str:
        return "<tr class='top-level-drag-row'><td colspan='6'>&nbsp;</td></tr>"

    # Options
    ##########################################################################

    def _showOptions(self, did: str) -> None:
        m = QMenu(self.mw)
        a = m.addAction(tr.actions_rename())
        assert a is not None
        qconnect(a.triggered, lambda b, did=did: self._rename(DeckId(int(did))))
        a = m.addAction(tr.actions_options())
        assert a is not None
        qconnect(a.triggered, lambda b, did=did: self._options(DeckId(int(did))))
        a = m.addAction(tr.actions_export())
        assert a is not None
        qconnect(a.triggered, lambda b, did=did: self._export(DeckId(int(did))))
        a = m.addAction(tr.actions_delete())
        assert a is not None
        qconnect(a.triggered, lambda b, did=did: self._delete(DeckId(int(did))))
        gui_hooks.deck_browser_will_show_options_menu(m, int(did))
        m.popup(QCursor.pos())

    def _export(self, did: DeckId) -> None:
        self.mw.onExport(did=did)

    def _rename(self, did: DeckId) -> None:
        def prompt(name: str) -> None:
            new_name = getOnlyText(
                tr.decks_new_deck_name(), default=name, title=tr.actions_rename()
            )
            if not new_name or new_name == name:
                return
            else:
                rename_deck(
                    parent=self.mw, deck_id=did, new_name=new_name
                ).run_in_background()

        QueryOp(
            parent=self.mw, op=lambda col: col.decks.name(did), success=prompt
        ).run_in_background()

    def _options(self, did: DeckId) -> None:
        display_options_for_deck_id(did)

    def _collapse(self, did: DeckId) -> None:
        node = self.mw.col.decks.find_deck_in_tree(self._render_data.tree, did)
        if node:
            node.collapsed = not node.collapsed
            set_deck_collapsed(
                parent=self.mw,
                deck_id=did,
                collapsed=node.collapsed,
                scope=DeckCollapseScope.REVIEWER,
            ).run_in_background()
            self._renderPage(reuse=True)

    def _handle_drag_and_drop(self, source: DeckId, target: DeckId) -> None:
        reparent_decks(
            parent=self.mw, deck_ids=[source], new_parent=target
        ).run_in_background()

    def _delete(self, did: DeckId) -> None:
        deck = self.mw.col.decks.find_deck_in_tree(self._render_data.tree, did)
        assert deck is not None
        deck_name = deck.name
        remove_decks(
            parent=self.mw, deck_ids=[did], deck_name=deck_name
        ).run_in_background()

    # Top buttons
    ######################################################################

    drawLinks = [
        ["", "shared", tr.decks_get_shared()],
        ["", "create", tr.decks_create_deck()],
        ["Ctrl+Shift+I", "import", tr.decks_import_file()],
    ]

    def _drawButtons(self) -> None:
        buf = ""
        drawLinks = deepcopy(self.drawLinks)
        for b in drawLinks:
            if b[0]:
                b[0] = tr.actions_shortcut_key(val=shortcut(b[0]))
            buf += """
<button title='%s' onclick='pycmd(\"%s\");'>%s</button>""" % tuple(b)
        self.bottom.draw(
            buf=buf,
            link_handler=self._linkHandler,
            web_context=DeckBrowserBottomBar(self),
        )

    def _onShared(self) -> None:
        openLink(f"{aqt.appShared}decks/")

    def _on_create(self) -> None:
        if op := add_deck_dialog(
            parent=self.mw, default_text=self.mw.col.decks.current()["name"]
        ):
            op.run_in_background()

    ######################################################################

    def _v1_upgrade_message(self, required: bool) -> str:
        if not required:
            return ""

        update_required = tr.scheduling_update_required().replace("V2", "v3")

        return f"""
<center>
<div class=callout>
    <div>
      {update_required}
    </div>
    <div>
      <button onclick='pycmd("v2upgrade")'>
        {tr.scheduling_update_button()}
      </button>
      <button onclick='pycmd("v2upgradeinfo")'>
        {tr.scheduling_update_more_info_button()}
      </button>
    </div>
</div>
</center>
"""

    def _confirm_upgrade(self) -> None:
        if self.mw.col.sched_ver() == 1:
            self.mw.col.mod_schema(check=True)
            self.mw.col.upgrade_to_v2_scheduler()
        self.mw.col.set_v3_scheduler(True)

        showInfo(tr.scheduling_update_done())
        self.refresh()



# Form implementation generated from reading ui file 'qt/aqt/forms/about.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_About(object):
    def setupUi(self, About):
        About.setObjectName("About")
        About.resize(410, 664)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(About.sizePolicy().hasHeightForWidth())
        About.setSizePolicy(sizePolicy)
        self.vboxlayout = QtWidgets.QVBoxLayout(About)
        self.vboxlayout.setContentsMargins(0, 0, 0, 0)
        self.vboxlayout.setObjectName("vboxlayout")
        self.label = AnkiWebView(parent=About)
        self.label.setProperty("url", QtCore.QUrl("about:blank"))
        self.label.setObjectName("label")
        self.vboxlayout.addWidget(self.label)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=About)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.vboxlayout.addWidget(self.buttonBox)

        self.retranslateUi(About)
        self.buttonBox.accepted.connect(About.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(About.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(About)

    def retranslateUi(self, About):
        _translate = QtCore.QCoreApplication.translate
        About.setWindowTitle(tr.about_about_anki())
from aqt.webview import AnkiWebView


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""Helpers for serializing third-party collections to a common JSON form."""

from __future__ import annotations

import json
from dataclasses import asdict, dataclass, field
from typing import Union

from anki.consts import STARTING_FACTOR_FRACTION
from anki.decks import DeckId
from anki.models import NotetypeId


@dataclass
class ForeignCardType:
    name: str
    qfmt: str
    afmt: str

    @staticmethod
    def front_back() -> ForeignCardType:
        return ForeignCardType(
            "Card 1",
            qfmt="{{Front}}",
            afmt="{{FrontSide}}\n\n<hr id=answer>\n\n{{Back}}",
        )

    @staticmethod
    def back_front() -> ForeignCardType:
        return ForeignCardType(
            "Card 2",
            qfmt="{{Back}}",
            afmt="{{FrontSide}}\n\n<hr id=answer>\n\n{{Front}}",
        )

    @staticmethod
    def cloze() -> ForeignCardType:
        return ForeignCardType(
            "Cloze", qfmt="{{cloze:Text}}", afmt="{{cloze:Text}}<br>\n{{Back Extra}}"
        )


@dataclass
class ForeignNotetype:
    name: str
    fields: list[str]
    templates: list[ForeignCardType]
    is_cloze: bool = False

    @staticmethod
    def basic(name: str) -> ForeignNotetype:
        return ForeignNotetype(name, ["Front", "Back"], [ForeignCardType.front_back()])

    @staticmethod
    def basic_reverse(name: str) -> ForeignNotetype:
        return ForeignNotetype(
            name,
            ["Front", "Back"],
            [ForeignCardType.front_back(), ForeignCardType.back_front()],
        )

    @staticmethod
    def cloze(name: str) -> ForeignNotetype:
        return ForeignNotetype(
            name, ["Text", "Back Extra"], [ForeignCardType.cloze()], is_cloze=True
        )


@dataclass
class ForeignCard:
    """Data for creating an Anki card.

    Usually a review card, as the default card generation routine will take care
    of missing new cards.

    due          --  UNIX timestamp
    interval     --  days
    ease_factor  --  decimal fraction (2.5 corresponds to default ease)
    """

    # TODO: support new and learning cards?
    due: int = 0
    interval: int = 1
    ease_factor: float = STARTING_FACTOR_FRACTION
    reps: int = 0
    lapses: int = 0


@dataclass
class ForeignNote:
    fields: list[str] = field(default_factory=list)
    tags: list[str] = field(default_factory=list)
    notetype: str | NotetypeId = ""
    deck: str | DeckId = ""
    cards: list[ForeignCard] = field(default_factory=list)


@dataclass
class ForeignData:
    notes: list[ForeignNote] = field(default_factory=list)
    notetypes: list[ForeignNotetype] = field(default_factory=list)
    default_deck: str | DeckId = ""

    def serialize(self) -> str:
        return json.dumps(self, cls=ForeignDataEncoder, separators=(",", ":"))


class ForeignDataEncoder(json.JSONEncoder):
    def default(self, obj: object) -> dict:
        if isinstance(
            obj,
            (ForeignData, ForeignNote, ForeignCard, ForeignNotetype, ForeignCardType),
        ):
            return asdict(obj)
        return json.JSONEncoder.default(self, obj)
